# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Compiler.
#
# The Infinity Note Compiler is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Infinity Note Compiler is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the Infinity Note Compiler.  If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from . import logger
from . import operations
from .types import INTTYPE
import inspect

# The primary goal of these optimizations is to reduce the instruction
# count, to aid consumers using interpreters to execute notes.  The
# secondary goal of these optimizations is to reduce the size of the
# bytecode in cases where this does not conflict with the primary goal
# of reducing instruction count.

class Optimizer(object):
    """Base class for all optimizers.
    """

    def visit_toplevel(self, toplevel):
        for node in toplevel.functions:
            node.accept(self)

    def debug_print_hit(self, location):
        if self.debug_print.is_enabled:
            optimization = inspect.stack()[1][0].f_code.co_name
            for prefix in ("__", "try_"):
                if optimization.startswith(prefix):
                    optimization = optimization[len(prefix):]
            self.debug_print("%s: %s\n" % (location.fileline, optimization))

class BlockOptimizer(Optimizer):
    """Optimizations performed before serialization.
    """

    debug_print = logger.debug_printer_for("blockopt")

    def visit_function(self, function):
        self.visited = {}
        function.entry_block.accept(self)
        self.try_combine_blocks(self.visited.keys())

    def try_combine_blocks(self, blocks):
        blocks = sorted((block.index, block) for block in blocks)
        blocks = [block for index, block in blocks]
        while self.__try_combine_blocks(blocks):
            pass

    def __try_combine_blocks(self, blocks):
        for i in range(len(blocks)):
            block_1 = blocks[i]
            for j in range(i + 1, len(blocks)):
                block_2 = blocks[j]
                if block_2.is_equivalent_to(block_1):
                    self.debug_print(
                        ("Blocks #%d and #%d are equivalent, "
                         + "removing #%d\n") % (block_1.index,
                                                block_2.index,
                                                block_2.index))
                    for block in blocks:
                        block.replace_exit(block_2, block_1)
                    blocks.remove(block_2)
                    return True
        return False

    def visit_basicblock(self, block):
        if self.visited.get(block, False):
            return
        self.visited[block] = True
        self.try_all_optimizations(block)
        if self.debug_print.is_enabled:
            self.debug_print(str(block) + "\n\n")
        for block in block.exits:
            block.accept(self)

    def try_all_optimizations(self, block):
        self.try_eliminate_cmp_bra_const_const(block)
        self.try_eliminate_lit0_cmp_before_bra(block)
        self.try_reverse_branch_exits(block)
        self.try_peephole(block, self.try_eliminate_identity_math, 2)
        self.try_peephole(block, self.try_use_plus_uconst, 2)

    def __tecbcc_helper(self, block):
        """Helper for try_eliminate_cmp_bra_const_const.
        """
        if len(block.entries) != 1:
            return
        if len(block.ops) < 2:
            return
        if not block.ops[0].is_load_constant:
            return
        constant = block.ops[0]
        if constant.type.basetype != INTTYPE:
            return
        return constant.value

    def try_eliminate_cmp_bra_const_const(self, block):
        """Optimize cases where the blocks following a conditional
        branch load the constants that the comparison pushed to the
        stack.

        This is relevant for libpthread notes.  All the libthread_db
        functions that the libpthread notes replace return a td_err_e
        error code defined as:

          typedef enum {
            TD_OK,  /* No error.  */
            TD_ERR, /* General error.  */
            ...     /* Specific errors.  */
          } td_err_e;

        Some libthread_db functions call proc_service functions which
        return a similar ps_err_e error code:

          typedef enum {
            PS_OK,  /* No error.  */
            PS_ERR, /* General error.  */
            ...     /* Specific errors.  */
          } ps_err_e;

        Note that TD_OK == PS_OK == 0 and TD_ERR == PS_ERR == 1.

        This optimizer replaces code of the following pattern:

            call        /* Some proc_service function.  */
            load PS_OK  /* == 0 */
            bne fail
            load TD_OK  /* == 0 */
            return
          fail:
            load TD_ERR /* == 1 */
            return

        With this:

            call        /* Some proc_service function.  */
            load PS_OK
            ne
        """
        # Does the block end with "comparison, branch"?
        if len(block.ops) < 2:
            return
        if not block.ops[-1].is_branch:
            return
        if not block.ops[-2].is_comparison:
            return

        # Do the successors start with "const 0" and "const 1"?
        constants = list(map(
                self.__tecbcc_helper, (block.nobranch_exit,
                                       block.branched_exit)))
        if 0 not in constants or 1 not in constants:
            return

        # Are the successors otherwise the same?
        s1, s2 = block.exits
        if s1.exits != s2.exits:
            return
        if len(s1.ops) != len(s2.ops):
            return
        for op1, op2 in list(zip(s1.ops, s2.ops))[1:-1]:
            if not op1.is_equivalent_to(op2):
                return

        self.debug_print_hit(block.ops[-1])

        # Reverse the comparison if necessary
        if constants == [1, 0]:
            block.ops[-2].reverse()

        # Lose one of the successor blocks (doesn't matter which)
        dead_block = block.exits.pop()
        dead_block.entries.remove(block)
        assert not dead_block.entries

        # Reduce the branch to a goto
        block.ops[-1] = operations.SyntheticGoto(block.ops[-1])

        # Move the the remaining successor and drop the ConstOp.
        # This messes with the types a bit (what was an INTTYPE
        # is now a BOOLTYPE) but that doesn't matter once it's
        # bytecode.
        [block] = block.exits
        removed_op = block.ops.pop(0)
        assert removed_op.is_load_constant

    def try_eliminate_lit0_cmp_before_bra(self, block):
        # Does the block end with "load 0, {eq,ne}, branch"?
        if len(block.ops) < 3:
            return
        if not block.ops[-1].is_branch:
            return
        if not block.ops[-2].is_comparison:
            return
        if block.ops[-2].dwarfname not in ("eq", "ne"):
            return
        if not block.ops[-3].is_load_constant:
            return
        if block.ops[-3].value != 0:
            return

        self.debug_print_hit(block.ops[-2])

        # Reverse the branch if necessary
        if block.ops[-2].dwarfname == "eq":
            block.exits.reverse()

        # Remove the load and the comparison
        removed_op = block.ops.pop(-3)
        assert removed_op.is_load_constant
        removed_op = block.ops.pop(-2)
        assert removed_op.is_comparison

    def try_reverse_branch_exits(self, block):
        # Does the block end with "compare, branch"?
        if len(block.ops) < 2:
            return
        if not block.ops[-1].is_branch:
            return
        if not block.ops[-2].is_comparison:
            return

        # Does the nobranch case immediately jump somewhere?
        tmp = block.nobranch_exit.first_op
        if not (tmp.is_goto or tmp.is_return):
            return

        # Does the branch case NOT immediately jump somewhere?
        tmp = block.branched_exit.first_op
        if tmp.is_goto or tmp.is_return:
            return

        self.debug_print_hit(block.ops[-2])

        # Reverse both the comparison and the branch
        block.ops[-2].reverse()
        block.exits.reverse()

    def try_peephole(self, block, action, size):
        start = 0
        while True:
            start = self.__try_peephole(block, action, size)
            if start is None:
                break

    def __try_peephole(self, block, action, size):
        """Helper for try_peephole.
        """
        for index in range(len(block.ops) - size):
            if action(block, index):
                return index

    IDENTITIES = {
        "plus": 0, "minus": 0, "mul": 1,
        "div": 1, "shl": 0, "shr": 0,
        "shra": 0, "or": 0, "xor": 0}

    def try_eliminate_identity_math(self, block, index):
        if not block.ops[index].is_load_constant:
            return False
        opname = getattr(block.ops[index + 1], "dwarfname", None)
        if opname is None:
            return False
        identity = self.IDENTITIES.get(opname, None)
        if identity is None:
            return False
        if block.ops[index].value != identity:
            return False

        self.debug_print_hit(block.ops[index + 1])

        # Remove the operations
        removed_op = block.ops.pop(index + 1)
        assert removed_op.dwarfname == opname
        removed_op = block.ops.pop(index)
        assert removed_op.is_load_constant

        return True

    def try_use_plus_uconst(self, block, index):
        if not block.ops[index].is_load_constant:
            return False
        if block.ops[index].value < 0:
            return False
        if not block.ops[index + 1].is_add:
            return False

        self.debug_print_hit(block.ops[index])

        # Insert the plus_uconst
        block.ops[index] = operations.PlusUConst(block.ops[index])

        # Remove the add
        removed_op = block.ops.pop(index + 1)
        assert removed_op.is_add

        return True

class StreamOptimizer(Optimizer):
    """Optimizations performed after serialization.
    """

    debug_print = logger.debug_printer_for("streamopt")

    def debug_print_stream(self, stream):
        self.debug_print("%s\n" % stream)

    def visit_function(self, function):
        function.ops.accept(self)

    def visit_operationstream(self, stream):
        while True:
            if self.try_remove_multijump(stream):
                continue
            if self.try_remove_goto_next(stream):
                continue
            if self.try_remove_unreachable_code(stream):
                continue
            break

    def try_remove_multijump(self, stream):
        for index, op in stream.items():
            target = stream.jumps.get(op, None)
            if target is None:
                continue
            if not target.is_goto:
                continue

            self.debug_print_hit(op)
            stream.retarget_jump(op, stream.jumps[target])
            self.debug_print_stream(stream)
            return True
        return False

    def try_remove_goto_next(self, stream):
        for index, op in stream.items():
            if index + 1 == len(stream.ops):
                continue
            if not op.is_goto:
                continue
            if stream.labels.get(op, None) is not None:
                continue
            if stream.jumps[op] is not stream.ops[index + 1]:
                continue

            self.debug_print_hit(op)
            stream.remove_by_index_op(index, op)
            self.debug_print_stream(stream)
            return True
        return False

    def try_remove_unreachable_code(self, stream):
        last_was_goto = False
        for index, op in stream.items():
            if last_was_goto:
                if stream.labels.get(op, None) is None:
                    self.debug_print_hit(op)
                    stream.remove_by_index_op(index, op)
                    self.debug_print_stream(stream)
                    return True
            last_was_goto = op.is_goto
        return False



# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Compiler.
#
# The Infinity Note Compiler is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Infinity Note Compiler is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the Infinity Note Compiler.  If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
# Python2 distutils can't cope with __future__.unicode_literals

from setuptools import find_packages, setup
from codecs import open
from os import path
import sys

here = path.realpath(path.dirname(__file__))

with open(path.join(here, "README.rst"), encoding="utf-8") as fp:
    long_description = fp.read()

# Remember to update requirements.txt
install_requires = [
    "arpy",
    "pyelftools >= 0.24",
]

# Ensure we have a suitable unittest
try:
    import unittest2 as unittest
except ImportError:
    import unittest
if not hasattr(unittest.TestCase, "assertIsInstance"):
    install_requires.append("unittest2")

# Ensure we don't ever release packages requiring unittest2
if "unittest2" in install_requires:
    for arg in sys.argv[1:]:
        for bail in ("register", "upload", "dist"):
            if arg.find(bail) >= 0:
                print("Don't use Python %s.%s for ‘%s’"
                      % (sys.version_info[:2] + (arg,)))
                sys.exit(1)

setup(
    name="i8c",
    version="0.0.4",
    description="Infinity Note Compiler",
    long_description=long_description,
    license="GPLv3+ and LGPLv2.1+",
    author="Gary Benson",
    author_email="infinity@sourceware.org",
    url="https://infinitynotes.org/",
    classifiers=[
        # https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved" +
            " :: GNU General Public License v3 or later (GPLv3+)",
        "License :: OSI Approved" +
            " :: GNU Lesser General Public License v2 or later (LGPLv2+)",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Compilers",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
    ],
    packages=find_packages("src"),
    package_dir = {"": "src"},
    install_requires=install_requires,
    entry_points={"console_scripts": ["i8c = i8c.compiler:main",
                                      "i8x = i8c.runtime:main"]},
    tests_require=["nose"],
    test_suite="nose.collector")



# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Execution Environment.
#
# The Infinity Note Execution Environment is free software; you can
# redistribute it and/or modify it under the terms of the GNU Lesser
# General Public License as published by the Free Software Foundation;
# either version 2.1 of the License, or (at your option) any later
# version.
#
# The Infinity Note Execution Environment is distributed in the hope
# that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with the Infinity Note Execution Environment; if not,
# see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from .. import constants
from . import CorruptNoteError, UnhandledNoteError

class Type(object):
    @classmethod
    def initialize(cls):
        cls.types = {}
        for item in globals().values():
            code = getattr(item, "code", None)
            if code is not None:
                assert code not in cls.types
                cls.types[code] = item

class BasicType(Type):
    def __init__(self):
        raise AssertionError

    @classmethod
    def pack(cls):
        return cls.code

    @classmethod
    def unpack(cls, input, start):
        return start + 1

class IntPtrType(BasicType):
    name = "intptr"

class IntegerType(IntPtrType):
    code = constants.I8_TYPE_INT

class PointerType(IntPtrType):
    code = constants.I8_TYPE_PTR

class OpaqueType(BasicType):
    code = constants.I8_TYPE_OPAQUE
    name = "opaque"

class FunctionType(Type):
    code = constants.I8_TYPE_FUNC
    name = "function"

    def __init__(self, ptypes=None, rtypes=None):
        assert (ptypes is None) == (rtypes is None)
        self.ptypes = ptypes
        self.rtypes = rtypes

    def __eq__(self, other):
        return not (self != other)

    def __ne__(self, other):
        assert self.ptypes is not None
        assert self.rtypes is not None
        assert other.ptypes is not None
        assert other.rtypes is not None
        return self.ptypes != other.ptypes or self.rtypes != other.rtypes

    def pack(self):
        return "%s%s(%s)" % (
            self.code, encode(self.rtypes), encode(self.ptypes))

    def unpack(self, input, start):
        assert self.ptypes is None and self.rtypes is None
        start, self.rtypes = decode(input, start + 1, "(")
        start, self.ptypes = decode(input, start + 1, ")")
        return start + 1

Type.initialize()

def decode(input, start=0, stop=None):
    result = []
    while start < len(input):
        code = input[start]
        if code == stop:
            return start, result
        thetype = Type.types.get(code, None)
        if thetype is None:
            raise UnhandledNoteError(input)
        if thetype is FunctionType:
            thetype = thetype()
        start = thetype.unpack(input, start)
        result.append(thetype)
    if stop is not None:
        raise CorruptNoteError(input)
    return result

def encode(list):
    return "".join([type.pack() for type in list])



# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Compiler.
#
# The Infinity Note Compiler is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Infinity Note Compiler is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the Infinity Note Compiler.  If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from tests import TestCase
from i8c.runtime import NoteError
from i8c.runtime import types

class TestTypeMixin(object):
    def __eq__(self, other):
        return isinstance(self, type(other))

    def __ne__(self, other):
        raise NotImplementedError

I, P, O, F = (types.IntegerType, types.PointerType,
              types.OpaqueType, types.FunctionType)

GOODENC = (("", []),
           ("i", [I]),
           ("p", [P]),
           ("o", [O]),
           ("iipiop", [I, I, P, I, O, P]),
           ("F()", [F([], [])]),
           ("Fi()", [F([], [I])]),
           ("F(p)", [F([P], [])]),
           ("Fo(op)", [F([O, P], [O])]),
           ("Fo(io)i", [F([I, O], [O]), I]),
           ("ooFo(io)ip", [O, O, F([I, O], [O]), I, P]),
           ("FioFpo(ii)(op)i",
            [F([O, P],
               [I, O, F([I, I],
                        [P, O])]),
             I]),
           ("poFioFpo(ii)p(oFi(o)p)i",
            [P, O, F(
                [O, F([O],
                      [I]), P],
                [I, O, F([I, I],
                         [P, O]), P]), I]),
           )

BADENC = (
    # Types only valid for externals
    "f", "x",

    # Uppercase versions of valid types
    "P", "I", "O",

    # Undefined types
    "q", "%",

    # Various bad function encodings
    "F",
    "F(",
    "Fii(pp",
    "F)",
    "Fp)i(op)",
    )

class FakeSlice(object):
    filename = "<testcase>"
    start = 0

    def __init__(self, bytes):
        self.bytes = bytes

    def __len__(self):
        return len(self.bytes)

    def __getitem__(self, key):
        return self.bytes[key]

class TestRuntimeTypes(TestCase):
    def test_good_encodings(self):
        """Check that the runtime can decode lists of types."""
        for encoded, expect in GOODENC:
            decoded = types.decode(FakeSlice(encoded))
            self.assertEqual(decoded, expect)
            self.assertEqual(types.encode(decoded), encoded)

    def test_bad_encodings(self):
        """Check that the runtime rejects bad encoded types lists."""
        for encoded in BADENC:
            self.assertRaises(NoteError, types.decode, FakeSlice(encoded))



# -*- coding: utf-8 -*-
# Copyright (C) 2016 Red Hat, Inc.
# This file is part of the Infinity Note Compiler.
#
# The Infinity Note Compiler is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Infinity Note Compiler is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the Infinity Note Compiler.  If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

# Export notes built by the testsuite for libi8x's corpus.

import hashlib
import os
import subprocess
import sys

topdir = os.path.dirname(os.path.dirname(os.path.realpath(sys.argv[0])))
outdir = os.path.join(topdir, "tests", "output")
subprocess.check_call(("/bin/rm", "-rf", outdir))
env = os.environ.copy()
env.update({"LIBI8X_TESTNOTE_EXPORT": "1"})
subprocess.check_call((sys.executable, "setup.py", "test"), env=env)
outdir = os.path.join(outdir, "for-libi8x")

filenames = []
for dirname, dirs, files in os.walk(outdir):
    for filename in files:
        filenames.append(os.path.join(dirname, filename))

# Don't bundle multiple copies of the same bytes.
by_md5sum = {}
strip = len(outdir + os.sep)
for filename in sorted(filenames):
    hash = hashlib.md5()
    with open(filename, "rb") as fp:
        block = fp.read(4096)
        if not block:
            break
        hash.update(block)
    md5sum = hash.hexdigest()
    if md5sum not in by_md5sum:
        by_md5sum[md5sum] = filename[strip:]

outfile = os.path.join(topdir, "notes-for-libi8x.tar")
subprocess.check_call(("/bin/tar", "-cf", outfile, "-C", outdir)
                      + tuple(by_md5sum.values()))



# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Compiler.
#
# The Infinity Note Compiler is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Infinity Note Compiler is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the Infinity Note Compiler.  If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from tests import TestCase
from i8c.compiler import loggers
from i8c.compiler import parser
import io
import sys

SOURCE = """\
define test::factorial returns int
    argument int x

    dup
    load 1
    bgt not_done_yet
    load 1
    return

not_done_yet:
    dup
    load 1
    sub
label: // ensure we hit the 1-exit case in blocks.Block.__str_
    load factorial
    call
    mul
"""

LAST_LOAD_NODE = """\
LoadOp
  Integer: 1 [int]"""

class TestDebugCode(TestCase):
    """Test various bits of debugging code."""

    def setUp(self):
        self.disable_loggers()
        self.saved_stderr = sys.stderr

    def tearDown(self):
        self.disable_loggers()
        sys.stderr = self.saved_stderr

    def test_loggers(self):
        """Exercise all the debug printers."""
        for logger in loggers.values():
            logger.enable()
        sys.stderr = io.StringIO()
        self.compile(SOURCE)

    def test_str_methods(self):
        """Check various __str__ methods."""
        tree, output = self.compile(SOURCE)
        # lexer.Token.__str__
        func = tree.one_child(parser.Function)
        ops = func.one_child(parser.Operations)
        token = ops.children[0].tokens[0]
        self.assertEqual(str(token), "<testcase>:5: 'dup'")
        # parser.TreeNode.__str__ with an annotated type
        load = list(ops.some_children(parser.LoadOp))[0]
        self.assertEqual(str(load), LAST_LOAD_NODE)



# -*- coding: utf-8 -*-
# Copyright (C) 2016 Red Hat, Inc.
# This file is part of the Infinity Note Compiler.
#
# The Infinity Note Compiler is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Infinity Note Compiler is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the Infinity Note Compiler.  If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os

class Variable(object):
    def __init__(self, name, default):
        self.as_str = os.environ.get(name, default)
        self.as_list = self.as_str.split()

    def __str__(self):
        return self.as_str

    def __add__(self, other):
        assert isinstance(other, list)
        return self.as_list + other


# Program for compiling C programs.
I8C_CC = Variable("I8C_CC", "gcc")

# Program for running the C preprocessor, with results to standard output.
I8C_CPP = Variable("I8C_CPP", "%s -E -x assembler-with-cpp" % I8C_CC)

# Program for compiling assembly files.
I8C_AS = Variable("I8C_AS", "%s -x assembler-with-cpp" % I8C_CC)



# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Compiler.
#
# The Infinity Note Compiler is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Infinity Note Compiler is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the Infinity Note Compiler.  If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from tests import TestCase

SOURCE = """\
define test::dup_over_equiv returns int
    argument int x
    argument int y
    argument int z

    blt 5, label

    load 0
    %s
    return

label:
    load 1
    pick %d
"""

class TestDupOverEquiv(TestCase):
    def test_dup_over_equiv(self):
        """Check that dup and over are equivalent to picks."""
        for equiv, index in (("dup", 0), ("over", 1)):
            tree, output = self.compile(SOURCE % (equiv, index))
            self.assertEqual(["lit5", "lt", equiv], output.opnames)



# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Execution Environment.
#
# The Infinity Note Execution Environment is free software; you can
# redistribute it and/or modify it under the terms of the GNU Lesser
# General Public License as published by the Free Software Foundation;
# either version 2.1 of the License, or (at your option) any later
# version.
#
# The Infinity Note Execution Environment is distributed in the hope
# that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with the Infinity Note Execution Environment; if not,
# see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from . import version

def version_message_for(program, license):
    license_txt, license_url = license
    license_url = "<%s>" % license_url
    return """\
%s %s
Copyright (C) 2015-16 Red Hat, Inc.
License %s\n  %s
This is free software: you are free to change and redistribute it.
There is NO WARRANTY, to the extent permitted by law.""" % (
        (program, version(), license_txt, license_url))

def usage_message_footer_for(program):
    return """

Report %s bugs to infinity@sourceware.org
Infinity home page: <https://infinitynotes.org/>""" % program



# -*- coding: utf-8 -*-
# Copyright (C) 2016 Red Hat, Inc.
# This file is part of the Infinity Note Compiler.
#
# The Infinity Note Compiler is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Infinity Note Compiler is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the Infinity Note Compiler.  If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from tests import TestCase

# XXX what?

SOURCE = """\
define test::i8xfail_0002 returns int
    extern ptr sym1
    deref sym1, uint32_t
"""

class TestI8XFail0002(TestCase):
    def test_i8xfail_0002(self):
        """Miscellaneous I8X failure #0002 check"""
        tree, output = self.compile(SOURCE)
        EXPECT = 0xdeadbeef
        with self.memory.builder() as mem:
            sym1 = mem.alloc("sym1")
            sym1.store_u32(0, EXPECT)
        self.assertEqual(output.call("test::i8xfail_0002()i"), [EXPECT])
